# -*- ruby -*-

#
# VFS for ruby - open anything with URI
# 
#   Copyright: Tatsuki Sugiura <sugi@nemui.org>
#   License: Ruby's original or GPL2
#   $LastChangedDate: 2003-11-20 09:16:53 +0900 (Thu, 20 Nov 2003) $
#   $LastChangedBy: sugi $
#

require 'vfs/common'
require 'vfs/file'
require 'vfs/dir'

Object.constants.member?("FileBuiltIn") or
  FileBuiltIn = File.clone
  
Object.constants.member?("DirBuiltIn") or
  DirBuiltIn = Dir.clone

Object.constants.member?("FileTestBuiltIn") or
  FileTestBuiltIn = FileTest.clone
  
module VFS
  module CompatLayer
    class Stat
      include Comparable

      def initialize(path)
        @vfsobj = VFS::File.new(path)
      end

      def <=>(other)
        @vfsobj.mtime <=> VFS::File.new(other).mtime
      end

      class_eval [
       :uid, :mtime, :writable?, :zero?, :blockdev?, :dev,
       :grpowned?, :rdev_major, :executable_real?, :setgid?, :nlink,
       :size, :atime, :readable_real?, :socket?, :blocks,
       :directory?, :owned?, :rdev, :executable?, :setuid?, :mode,
       :ftype, :readable?, :symlink?, :blksize, :size?, :gid, :ctime,
       :writable_real?, :chardev?, :ino, :pipe?, :rdev_minor, :file?,
       :sticky?
      ].map { |m| %{
        def #{m}
          @vfsobj.#{m}
        end
      } }.join

    end # Stat

    class File < VFS::File
      class << self
        def new(path, *args, &block)
          VFS::File.new(path).open(*args, &block)
        end
        alias_method :open, :new

        def umask(*mask)
          FileBuiltIn.umask(*mask)
        end

        def utime(n_atime, n_mtime, *files)
          files.each { |f|
            fo = VFS::File.new(f)
            fo.atime = n_atime
            fo.mtime = n_mtime
          }
          files.length
        end

        def chmod(mode, *files)
          files.each { |f|
            VFS::File.new(f).mode = mode
          }
          files.length
        end
        
        def lchmod(mode, *files)
          files.each { |f|
            VFS::File.new(f).lmode = mode
          }
          files.length
        end
        
        def chown(user, group, *files)
          files.each { |f|
            fo = VFS::File.new(f)
            fo.uid = user
            fo.gid = group
          }
          files.length
        end

        def lchown(user, group, *files)
          files.each { |f|
            fo = VFS::File.new(f)
            fo.luid = user
            fo.lgid = group
          }
          files.length
        end

        def unlink(*files)
          files.each { |f|
            VFS::File.new(f).unlink
          }
          files.length
        end 
        alias :delete :unlink

        def expand_path(path, base = nil)
          # FIXME!!
          if path =~ /^[a-z+_-]+:/
            require 'uri'
            ret = ""
            u = URI.parse(path)
            while u.scheme && u.opaque && u.opaque =~ /^[a-z+_-]+:/
              ret << u.scheme << ":"
              u = URI.parse(u.opaque)
            end
            u.path = FileBuiltIn.expand_path(u.path, "/")
            ret << u.to_s
          elsif base == nil
            base = DirBuiltIn.pwd
            FileBuiltIn.expand_path(path, base)
          end
        end

        def fnmatch(*args)
          FileBuiltIn.fnmatch(*args)
        end
        alias_method :fnmatch?, :fnmatch

        def stat(path)
          if VFS.scheme(path) == "file"
            FileBuiltIn.stat(path)
          else
            VFS::CompatLayer::Stat.new(path)
          end
        end

        def read(*args)
          FileBuiltIn.read(*args)
        end

        def readlines(*args)
          FileBuiltIn.readlines(*args)
        end

        def select(*args)
          FileBuiltIn.select(*args)
        end

#        undef_method :mtime=, :atime=, :uid, :dev, :rdev_major, :nlink,
#          :blocks, :rdev, :mode, :blksize, :gid, :ino, :rdev_minor

      end # class << self(CompatLayer::File)

      class_eval( "class << self;" +
        (FileBuiltIn.singleton_methods - self.singleton_methods).map {|m|
          %{
            def #{m}(*args)
              VFS::File.new(args.shift).#{m}(*args)
            end
          }
        }.join + "end")
      
    end  # CompatLayer::File

    class Dir < VFS::Dir
      class << self
        def chroot(path)
          DirBuiltIn.chroot(path)
        end

        def chdir(*args, &block)
          # FIXME. must support chdir on VFS
          DirBuiltIn.chdir(*args, &block)
        end
      end
      class_eval( "class << self;" +
        (DirBuiltIn.singleton_methods - self.singleton_methods).map {|m|
          %{
            def #{m}(*args)
              VFS::Dir.new(args.shift).#{m}(*args)
            end
          }
        }.join + "end")
    end # Dir

    module FileTest
      module_eval "module_function;" + [
        :writable?, :zero?, :blockdev?, :exist?, :grpowned?,
        :executable_real?, :setgid?, :readable_real?, :size, :socket?,
        :directory?, :owned?, :executable?, :setuid?, :readable?,
        :symlink?, :size?, :writable_real?, :chardev?, :exists?, :pipe?,
        :file?, :sticky?].map { |m| %{
          def #{m}(filename)
            VFS::File.new(filename).#{m}
          end
        }
      }.join
    end

    extend VFS::MethodGenerator

    @@permanently    = false
    @@current        = :BuiltIn
    @@file_m_diff    = []
    @@dir_m_diff     = []
    @@filetest_m_diff = []
    
    class_attr_accessor :permanently, :current

    class << self
      def allow_colon3_syntax?
        return false if ::VERSION.split(/\./)[0..2].join(".").to_f < 1.8
        return false if ::VERSION == "1.8.0" &&
                        ::RUBY_RELEASE_DATE.gsub(/-/, '').to_i < 20030814
        true
      end
      def allow_methods_with_flag?
        ::VERSION.split(/\./)[0..1].join(".").to_f < 1.7 ? false : true
      end

      def log(msg, *level)
	VFS::Log.log("VFS(CompatLayer): " + msg, *level)
      end

      def override
#	unless allow_methods_with_flag?
#	  return false
#	end
	if @@current != :BuiltIn
	  log("built-in class was already overrided? skip override.", Log::DEBUG)
	  return true
	end
	log("overriding built-in class.", Log::DEBUG)
	override_kernel_methods
	@@file_m_diff = gen_class_method_delegator(::File,
                                                   ::VFS::CompatLayer::File,
                                                   @@file_m_diff)
	@@filetest_m_diff = gen_class_method_delegator(::FileTest,
                                                       ::VFS::CompatLayer::FileTest,
                                                       @@filetest_m_diff)
	@@dir_m_diff  = gen_class_method_delegator(::Dir,
                                                   ::VFS::CompatLayer::Dir,
                                                   @@dir_m_diff)
	@@current = :VFS
      end

      def restore
#	unless allow_methods_with_flag?
#	  return false
#	end
	if @@current != :VFS
	  log("built-in class was not overrided? skip restore.", Log::DEBUG)
	  return true
	end
	if @@permanently
	  log("permanent flug is ture. skip restore.", Log::DEBUG)
	  return true
	end
	restore_kernel_methods
	@@file_m_diff = gen_class_method_delegator(::File, ::FileBuiltIn,
                                                   @@file_m_diff)
	@@filetest_m_diff = gen_class_method_delegator(::FileTest,
                                                       ::FileTestBuiltIn,
                                                       @@filetest_m_diff)
	@@dir_m_diff  = gen_class_method_delegator(::Dir,  ::DirBuiltIn,
                                                   @@dir_m_diff)
	@@current = :BuiltIn
      end

      def override_kernel_methods
	log("overriding kernel methods. go eval...", Log::DEBUG)
        ::Kernel.module_eval %Q{
          begin
            # why I can not use respond_to? :open_bulitin ???
            method :open_builtin
            # nothing to do
          rescue NameError
            alias_method :open_builtin, :open
            def open(*args, &block)
              VFS::CompatLayer::File.open(*args, &block)
            end
          end # rescue NameError
	}
      end

      def restore_kernel_methods
	log("restoring kernel methods. go eval...", Log::DEBUG)
        ::Kernel.module_eval %Q{
	    begin
	      # why I can not use respond_to? :open_bulitin ???
	      method :open_builtin
	      remove_method :open
	      alias_method  :open, :open_builtin
	      remove_method :open_builtin
	    rescue NameError
	      # nothing to do
	    end
	}
      end

      # return added methods
      def gen_class_method_delegator(target_class, class_new, del_methods=[])
	log("gen class methods for #{target_class} to #{class_new}.",
            Log::DEBUG)
        new_methods = class_new.singleton_methods
        old_methods = allow_methods_with_flag? ? 
          target_class.singleton_methods(false) :
          target_class.singleton_methods
        rem_methods = old_methods - new_methods - del_methods
        eval_str    = "class << self;" 
        eval_str   << (old_methods & new_methods).map { |m|
          "begin remove_method :#{m}; rescue NameError; end; "}.join("\n")

        eval_str << new_methods.map { |m|
	  if m.to_s =~ /=$/
            %{
              def #{m}(*args)
                ::#{class_new}.#{m} *args
              end
            }
          else
            %{
              def #{m}(*args, &block)
                ::#{class_new}.#{m}(*args, &block)
              end
            }
          end
        }.join

        if del_methods.length > 0
          eval_str << "remove_method " <<
            del_methods.map{|m| ":#{m}"}.join(",")  << "\n"
          VFS::Log.log("VFS(CompatLayer): following methods was deleted." +
		       " (#{del_methods.join(', ')})", VFS::Log::DEBUG)
        end
        if rem_methods.length > 0
          VFS::Log.log("VFS(CompatLayer): following methods was remaind." +
		       " (#{rem_methods.join(', ')})", VFS::Log::INFO)
        end

        eval_str << ";end"
        target_class.class_eval eval_str
        new_methods - old_methods
      end # gen_class_method_delegator

    end # class << self
  end # CompatLayer
end # VFS


# unless VFS::CompatLayer.allow_methods_with_flag?
#   # simple override is only for ruby 1.6 and older.
#   # use VFS::ConpatLayer above for 1.8 and later.
#   # but eval may be slow...
#   #
#   module Kernel
#     def open(*args, &block)
#       VFS::CompatLayer::File.open(*args, &block)
#     end
#   end
#   
#   class File
#     class << self
#       method_def = ""
#       (VFS::File.instance_methods(false) + [:new]).each {|m|
# 	m.to_s =~ /=$/ and next
# 	begin
# 	  remove_method m
# 	rescue NameError
# 	end
# 	class_eval %Q{
#           def #{m}(*args, &block)
# 	    VFS::CompatLayer::File.#{m}(*args, &block)
# 	  end
# 	}
#       }
#     end
#   end
#   
#   class Dir
#     class << self
#       (VFS::Dir.instance_methods(false) +[:new]).each {|m|
# 	m.to_s =~ /=$/ and next
# 	begin
# 	  remove_method m
# 	rescue NameError
# 	end
# 	class_eval %Q{
#           def #{m}(*args, &block)
# 	    VFS::CompatLayer::Dir.#{m}(*args, &block)
# 	  end
# 	}
#       }
#     end
#   end
# end
# 
